@@ -682,6 +682,8 @@ impl core::str::FromStr for Variation {
 pub trait TagExt {
     fn default_script() -> Self;
     fn default_language() -> Self;
+    fn to_lowercase(&self) -> Self;
+    fn to_uppercase(&self) -> Self;
 }
 
 impl TagExt for Tag {
@@ -694,4 +696,28 @@ impl TagExt for Tag {
     fn default_language() -> Self {
         Tag::from_bytes(b"dflt")
     }
+
+    /// Converts tag to lowercase.
+    #[inline]
+    fn to_lowercase(&self) -> Self {
+        let b = self.to_bytes();
+        Tag::from_bytes(&[
+            b[0].to_ascii_lowercase(),
+            b[1].to_ascii_lowercase(),
+            b[2].to_ascii_lowercase(),
+            b[3].to_ascii_lowercase(),
+        ])
+    }
+
+    /// Converts tag to uppercase.
+    #[inline]
+    fn to_uppercase(&self) -> Self {
+        let b = self.to_bytes();
+        Tag::from_bytes(&[
+            b[0].to_ascii_uppercase(),
+            b[1].to_ascii_uppercase(),
+            b[2].to_ascii_uppercase(),
+            b[3].to_ascii_uppercase(),
+        ])
+    }
 }
