@@ -3,7 +3,6 @@ A complete [harfbuzz](https://github.com/harfbuzz/harfbuzz) shaping algorithm po
 */
 
 #![no_std]
-#![doc(html_root_url = "https://docs.rs/rustybuzz/0.4.0")]
 #![warn(missing_docs)]
 
 extern crate alloc;
@@ -18,7 +17,6 @@ mod normalize;
 mod shape;
 mod plan;
 mod face;
-mod tables;
 mod tag;
 mod tag_table;
 mod text_parser;
